@@ -1,6 +1,6 @@
 package Net::OpenSSH::Compat::SSH2;
 
-our $VERSION = '0.02';
+our $VERSION = '0.08';
 
 use strict;
 use warnings;
@@ -13,9 +13,11 @@ use File::Basename ();
 use Fcntl ();
 use Carp ();
 use Scalar::Util ();
+use POSIX ();
 
 require Exporter;
-our @ISA = qw(Exporter Net::OpenSSH::Compat::SSH2::Base);
+our @ISA = qw(Exporter);
+our @CARP_NOT = qw(Net::OpenSSH);
 
 use Net::OpenSSH::Compat::SSH2::Constants;
 
@@ -52,7 +54,7 @@ sub import {
         }
     }
     __PACKAGE__->export_to_level(1, $class,
-                                 grep $_ ne ':supplant', @_));
+                                 grep $_ ne ':supplant', @_);
 }
 
 sub version { "1.2.6 (".__PACKAGE__."-$VERSION)" }
@@ -102,9 +104,7 @@ sub sock { undef }
 sub trace { }
 
 my @_auth_list = qw(publickey password);
-sub auth_list {
-    wantarray ? @_auth_list : join(',', @_auth_list);
-}
+sub auth_list { wantarray ? @_auth_list : join(',', @_auth_list) }
 
 sub connect {
     @_ < 2 and croak "Net::SSH2::connect: not enough parameters";
@@ -128,8 +128,6 @@ my %method_default = ( HOSTKEY => 'ssh-rsa',
                        COMP    => 'none',
                      );
 
-sub auth_list { wantarray ? qw(publickey password) : 'publickey,password' }
-
 sub method {
     my $cpt = shift;
     my $attr = shift;
@@ -253,6 +251,7 @@ sub blocking {
 
 sub scp_get {
     my ($cpt, $remote, $local) = @_;
+    $cpt->_check_state('ok');
     unless (defined $local) {
         $local = File::Basename::basename($remote);
     }
@@ -267,6 +266,7 @@ sub scp_get {
 
 sub scp_put {
     my ($cpt, $local, $remote) = @_;
+    $cpt->_check_state('ok');
     unless (defined $remote) {
         $remote = File::Basename::basename($local);
     }
@@ -281,6 +281,7 @@ sub scp_put {
 
 sub channel {
     my $cpt = shift;
+    $cpt->_check_state('ok');
     my $class = join('::', ref($cpt), 'Channel');
     my $chan = $class->_new($cpt);
     push @{$cpt->{channels}}, $chan;
@@ -290,6 +291,7 @@ sub channel {
 
 sub sftp {
     my $cpt = shift;
+    $cpt->_check_state('ok');
     my $class = join('::', ref($cpt), 'SFTP');
     $class->_new($cpt);
 }
@@ -310,7 +312,7 @@ sub _new {
 
 sub _hash { *{shift @_}{HASH} }
 
-sub _parent { shift->{hash}{cpt} }
+sub _parent { shift->_hash->{cpt} }
 
 sub ext_data { $_[0]->_hash->{ext_data} = $_[1] }
 
@@ -377,11 +379,9 @@ sub close {
     $chan->SUPER::close;
     $ch->{err} and close($ch->{err});
     # warn "reaping $ch->{pid}";
-    if (defined $ch->{pid}) {
-        waitpid $ch->{pid}, 0;
-        $ch->{exit_status} = $?;
-    }
+    $chan->_exit_status(1);
     $ch->{state} = 'closed';
+    $ch->{eof} = 1;
     1;
 }
 
@@ -408,12 +408,28 @@ sub wait_closed {
     $ch->{state} eq 'closed';
 }
 
-sub exit_status {
-    my $chan = shift;
-    $chan->_check_state('closed') or return;
-    $chan->_hash->{exit_status} >> 8;
+sub _exit_status {
+    my ($chan, $wait) = @_;
+    my $ch = $chan->_hash;
+    return $ch->{exit_status} if defined $ch->{exit_status};
+    return 0 unless defined $ch->{pid};
+    while (1) {
+        my $pid = waitpid $ch->{pid}, ($wait ? 0 : POSIX::WNOHANG());
+        if ($pid == $ch->{pid}) {
+            return $ch->{exit_status} = $?;
+        }
+        if ($pid < 0 and $! == Errno::ECHILD) {
+            return $ch->{exit_status} = 0;
+        }
+        return 0 unless $wait;
+        select(undef, undef, undef, 0.1);
+    }
 }
 
+sub exit_status { shift->_exit_status(0) >> 8 }
+
+sub exit_signal { shift->_exit_status(0) & 255 }
+
 sub blocking { shift->_hash->{cpt}->blocking(@_) }
 
 sub _blocking {
@@ -442,19 +458,30 @@ sub write {
 
 sub read {
     my ($chan, undef, $size, $ext) = @_;
+    $size ||= 1024;
     $chan->_check_state('exec') or return;
     my $ch = $chan->_hash;
-    my $fd;
     if ($ext) {
-        $fd = $ch->{err} or
+        my $fd = $ch->{err} or
             $chan->_set_error(LIBSSH2_ERROR_CHANNEL_UNKNOWN => 'no ext channel available');
+        return sysread($fd, $_[1], $size);
     }
     else {
-        $fd = $chan;
+        unless ($ch->{blocking}) {
+            my $fno = fileno($chan);
+            my $v = '';
+            vec($v, $fno, 1) = 1;
+            select($v, undef, undef, 0);
+            vec($v, $fno, 1) or return 0;
+        }
+        my $bytes = sysread($chan, $_[1], $size || 0);
+        $ch->{eof} = 1 unless $bytes;
+        return $bytes;
     }
-    sysread($fd, $_[1], $size || 0)
 }
 
+sub eof { shift->_hash->{eof} || 0}
+
 sub flush { 0 }
 
 package Net::OpenSSH::Compat::SSH2::SFTP;
@@ -725,7 +752,7 @@ change.
 
 The hash C<%Net::OpenSSH::Compat::SSH2::DEFAULTS> can be used to set
 default values for L<Net::OpenSSH> and other modules called under the
-hood and otherwise not accesible through the Net::SSH2 API.
+hood and otherwise not accessible through the Net::SSH2 API.
 
 The entries currently supported are:
 
@@ -758,7 +785,7 @@ Extra options passed to C<Net::SFTP::Foreign::new> constructor.
 B<This is a work in progress.>
 
 Besides that, there are some functionality of Net::SSH2 that can not
-be emulated with Net::OpenSSH. Fortunatelly, the missing bits are rarely
+be emulated with Net::OpenSSH. Fortunately, the missing bits are rarely
 used so probably you may not need them at all.
 
 Specifically, the return values from the C<$ssh2-E<gt>method($ATTR)>
@@ -806,7 +833,8 @@ upon: L<http://www.openssh.org/donations.html>.
 
 =head1 COPYRIGHT AND LICENSE
 
-Copyright (C) 2011 by Salvador FandiE<ntilde>o (sfandino@yahoo.com)
+Copyright (C) 2011, 2014-2016 by Salvador FandiE<ntilde>o
+(sfandino@yahoo.com)
 
 This library is free software; you can redistribute it and/or modify
 it under the same terms as Perl itself, either Perl version 5.10.0 or,
